



from typing import Any

from .expand import expand_tool_state
from .flatten import get_flattened_tool_state
from .resolve import resolve_values
from .standardisation import standardise_tool_state


# MODULE ENTRY
def load_tool_state(step: dict[str, Any]) -> dict[str, Any]:
    step['tool_state'] = expand_tool_state(step)
    step['tool_state'] = get_flattened_tool_state(step)
    step['tool_state'] = resolve_values(step)
    step['tool_state'] = standardise_tool_state(step)
    return step['tool_state']






from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

from .Param import Param


class ParamRegister:
    def __init__(self):
        self.params: list[Param] = []

    def list(self) -> list[Param]:
        return self.params
    
    def add(self, param: Param) -> None:
        """adds a param to register. enforces unique param var names"""
        self.params.append(param)

    def get(self, query: str, strategy: str='exact') -> Optional[Param]:
        """performs search using the specified search strategy"""
        strategy_map = {
            'exact': ExactSearchStrategy(),
            'lca': LCASearchStrategy(),
            'filepath': FilepathSearchStrategy(),
        }

        search_strategy = strategy_map[strategy]
        return search_strategy.search(query, self.params)



class SearchStrategy(ABC):    
    @abstractmethod
    def search(self, query: str, params: list[Param]) -> Optional[Param]:
        """searches for a param using some concrete strategy"""
        ...

class ExactSearchStrategy(SearchStrategy):
    def search(self, query: str, params: list[Param]) -> Optional[Param]:
        """searches for a param using param name"""
        for param in params:
            if param.name == query:
                return param
        return None

@dataclass
class LCAParam:
    split_name: list[str]
    param: Param

class LCASearchStrategy(SearchStrategy):
    def search(self, query: str, params: list[Param]) -> Optional[Param]:
        """searches for a param using LCA"""
        split_query = query.split('.')
        remaining_params = self.init_datastructure(params)

        for i in range(1, len(split_query) + 1):
            remaining_params = [p for p in remaining_params if len(p.split_name) >= i]
            remaining_params = [p for p in remaining_params if p.split_name[-i] == split_query[-i]]

        if len(remaining_params) > 0:
            # return shortest param name which matches the query
            remaining_params.sort(key=lambda x: len(x.split_name))
            return remaining_params[0].param

    def init_datastructure(self, params: list[Param]):
        return [LCAParam(param.name.split('.'), param) for param in params]
        

class FilepathSearchStrategy(SearchStrategy):
    def search(self, query: str, params: list[Param]) -> Optional[Param]:
        """
        searches for a param by matching the specified 
        from_work_dir path to a given filepath
        """
        raise NotImplementedError
        # for param in params.values():
        #     if hasattr(param, 'from_work_dir') and param.from_work_dir == query:
        #         return param




import os
import json
import xml.etree.ElementTree as et
from typing import Optional


def is_galaxy_workflow(path: str) -> bool:
    """checks that the provided path is a galaxy workflow"""
    with open(path, 'r') as fp:
        data = json.load(fp)
    if data['a_galaxy_workflow'] == 'true':
        return True
    return False

def is_tool_xml(path: str) -> bool:
    """checks that the provided path is tool xml"""
    root = et.parse(path).getroot()
    if root.tag == 'tool':
        return True
    return False

def is_macro_xml(path: str) -> bool:
    """checks that the provided path is tool xml"""
    root = et.parse(path).getroot()
    if root.tag == 'macros':
        return True
    return False

def get_xml_id(filepath: str) -> str:
    tree = et.parse(filepath)
    root = tree.getroot()
    return str(root.attrib['id']) # type: ignore

def get_xml_by_id(xmldir: str, query_id: str) -> Optional[str]:
    xmls = [x for x in os.listdir(xmldir) if x.endswith('.xml') and 'macros' not in x]
    for xml in xmls:
        filepath = f'{xmldir}/{xml}'
        tool_id = get_xml_id(filepath)
        if query_id == tool_id:
            return xml
    return None 

def get_macros(xmldir: str) -> list[str]:
    out: list[str] = []    
    xmls = [x for x in os.listdir(xmldir) if x.endswith('.xml')]
    for xml in xmls:
        path = f'{xmldir}/{xml}'
        if is_macro_xml(path):
            out.append(path)
    return out





import json
import filelock
from typing import Any

from .Container import Container


class ContainerCache:
    def __init__(self, cache_path: str):
        self.cache_path = cache_path

    def get(self, tool: str) -> list[Container]:
        """returns all containers with this tool id"""
        cache = self._load_cache()
        out: list[Container] = []
        if tool in cache:
            for details_dict in cache[tool]:
                out.append(Container(details_dict))
        return out

    def add(self, container: Container, override: bool=False):
        tool = container.tool_name
        if self.exists(container):
            return
        else:
            cache = self._load_cache()
            if tool not in cache:
                cache[tool] = []
            cache[tool].append(container.__dict__)
            self._write_cache(cache)

    def exists(self, container: Container) -> bool:
        cache = self._load_cache()
        if container.tool_name in cache:
            for details in cache[container.tool_name]:
                if details['url'] == container.url:
                    return True
        return False

    def _load_cache(self) -> dict[str, list[dict[str, str]]]:
        try:
            lockpath = f"{self.cache_path.rsplit('.', 1)[0]}.lock"
            lock = filelock.FileLock(lockpath)
            with lock.acquire(timeout=10):
                with open(self.cache_path, 'r') as fp:
                    return json.load(fp)
        except FileNotFoundError:
            return {}

    def _write_cache(self, cache: dict[str, Any]) -> None:
        filepath = self.cache_path
        lockpath = f"{filepath.rsplit('.', 1)[0]}.lock"
        lock = filelock.FileLock(lockpath)
        with lock.acquire(timeout=10):
            with open(self.cache_path, 'w') as fp:
                json.dump(cache, fp)






from abc import ABC, abstractmethod
from typing import Optional

from ..components import InputComponent
from ..components import Positional
from ..components import Flag
from ..components import Option


class ComponentOrderingStrategy(ABC):
    @abstractmethod
    def order(self, components: list[InputComponent], disallow: Optional[list[str]]=None) -> list[InputComponent]:
        """annotates components with command line positions"""
        ...

# TODO WTF epath cmd_pos orderings for components vs Command() orderings for components
class SimplifiedComponentOrderingStrategy(ComponentOrderingStrategy):
    def order(self, components: list[InputComponent], disallow: Optional[list[str]]=None) -> list[InputComponent]:
        start_pos_count = self.count_starting_positionals(components)
        self.assign_positional_positions(components, start_pos_count)
        self.assign_flag_option_positions(components, start_pos_count)
        return components

    def count_starting_positionals(self, components: list[InputComponent]) -> int:
        i: int = 0
        while i < len(components) and isinstance(components[i], Positional):
            i += 1
        return i 

    def assign_positional_positions(self, components: list[InputComponent], start_pos_count: int) -> None:
        positionals = [c for c in components if isinstance(c, Positional)]
        for i, comp in enumerate(positionals):
            if i >= start_pos_count:
                comp.cmd_pos = i + 1
            else:
                comp.cmd_pos = i
                comp.before_opts = True

    def assign_flag_option_positions(self, components: list[InputComponent], start_pos_count: int) -> None:
        options_flags = [c for c in components if isinstance(c, Option) or isinstance(c, Flag)]
        for comp in options_flags:
            comp.cmd_pos = start_pos_count


class RealisticComponentOrderingStrategy(ComponentOrderingStrategy):
    def order(self, components: list[InputComponent], disallow: Optional[list[str]]=None) -> list[InputComponent]:
        # wayyy more complicated
        # 
        pos: int = 0
        for component in components:
            component.cmd_pos = pos
            if isinstance(component, Positional):
                pos += 1
        return components








from dataclasses import dataclass
from typing import Tuple

from ..TextRender import TextRender
from ..tool.ToolInputText import ToolInputText
from .. import ordering

from galaxy2janis.gx.command.components import InputComponent
from galaxy2janis.gx.command.components import Flag
from galaxy2janis.gx.command.components import Option
from galaxy2janis.gx.command.components import Positional



@dataclass
class ToolInputSectionText(TextRender):
    def __init__(self, entity: list[InputComponent]):
        super().__init__()
        self.entity = entity

    @property
    def imports(self) -> list[Tuple[str, str]]:
        imports: list[Tuple[str, str]] = []
        for inp in self.entity:
            imports += ToolInputText(inp).imports
        imports = list(set(imports))
        return ordering.order_imports(imports)

    def render(self) -> str:
        out_str: str = ''

        out_str += 'inputs = ['

        out_str += '\n\t# Positionals\n'
        for positional in self._get_positionals():
            render = ToolInputText(positional)
            out_str += f'{render.render()},\n'

        out_str += '\n\t# Flags\n'
        for flag in self._get_flags():
            render = ToolInputText(flag)
            out_str += f'{render.render()},\n'

        out_str += '\n\t# Options\n'
        for option in self._get_options():
            render = ToolInputText(option)
            out_str += f'{render.render()},\n'
        
        out_str += '\n]\n'
        return out_str

    def _get_positionals(self) -> list[Positional]:
        positionals = [x for x in self.entity if isinstance(x, Positional)]
        return ordering.order_positionals(positionals)
    
    def _get_flags(self) -> list[Flag]:
        flags = [x for x in self.entity if isinstance(x, Flag)]
        return ordering.order_flags(flags)
    
    def _get_options(self) -> list[Option]:
        options = [x for x in self.entity if isinstance(x, Option)]
        return ordering.order_options(options)

    



from abc import ABC, abstractmethod
from typing import Any, Optional
from galaxy2janis.gx.gxtool.param import Param
from uuid import uuid4

from galaxy2janis import tags
from galaxy2janis import datatypes

 
class InputComponent(ABC):
    def __init__(self):
        self.uuid: str = str(uuid4())
        self.gxparam: Optional[Param] = None
        self.cmd_pos: int = -1
        self.forced_optionality: Optional[bool] = None
        self.forced_array: Optional[bool] = None
 
    @property
    @abstractmethod
    def name(self) -> str:
        """
        returns a name for this component. created depending on what
        information is available to the component
        """
        ...
    
    @property
    def tag(self) -> str:
        return tags.get(self.uuid)

    @property
    def datatype(self) -> datatypes.JanisDatatype:
        return datatypes.get(self)

    @property
    @abstractmethod
    def default_value(self) -> Any:
        """
        gets the default value of the component.
        uses the components observed values (positionals / options) and
        galaxy param information if available.
        """
        ...

    @property
    @abstractmethod
    def optional(self) -> bool:
        """
        returns whether the component is optional or not.
        uses galaxy param information if available, otherwise uses the presence array. flags are always optional
        """
        ...

    @property
    @abstractmethod
    def array(self) -> bool:
        """
        returns whether the component is an array or not
        uses galaxy param information if available.
        flags components are never arrays.
        """
        ...

    @property
    @abstractmethod
    def docstring(self) -> Optional[str]:
        """
        gets helptext for the component. uses galaxy param if available,
        otherwise usually just presents the witnessed values as examples. 
        """
        ...

    @abstractmethod
    def update(self, incoming: Any) -> None:
        """
        updates this component with information from another similar component. 
        also sets the component as being present in the cmdstr being parsed.
        example: 
        str1: abricate $input.fastq --db card > out.txt
        str2: abricate $input.fastq --db resfinder > out.txt
        the --db option should be updated so we know its possible values
        are (at least) card, resfinder
        """
        ...








from typing import Any, Optional

from ....gxtool.param.Param import Param
from ..ValueRecord import ValueRecord
from .InputComponent import InputComponent
from . import utils



class Positional(InputComponent):
    def __init__(self) -> None:
        super().__init__()
        self.before_opts: bool = False
        self.values: ValueRecord = ValueRecord()

    @property
    def name(self) -> str:
        # just return script if its a script
        if self.values.script:
            return 'script'
        # get name from galaxy param if available
        if self.gxparam:
            return self.gxparam.name  # what about adv.reference?
        # otherwise, most commonly witnessed option value as name
        pseudo_name = self.values.most_common_value
        if not pseudo_name:
            pseudo_name = 'positional'
        return pseudo_name.strip('$')

    @property
    def default_value(self) -> Any:
        """gets the default value for this component"""
        if self.gxparam:
            default = self.gxparam.default
        elif len(self.values.unique) == 1:
            default = self.values.unique[0]
        elif len(self.values.unique) > 1:
            if self.values.script:
                default = self.values.script
            elif self.values.env_var:
                default = self.values.env_var
            else:
                default = self.values.most_common_value
        else:
            default = None
        return utils.sanitise_default_value(default)

    @property
    def optional(self) -> bool:
        if self.forced_optionality is not None:
            return self.forced_optionality
        elif self.gxparam: 
            return self.gxparam.optional
        return False

    @property
    def array(self) -> bool:
        if self.forced_array is not None:
            return self.forced_array
        elif self.gxparam: 
            return self.gxparam.array
        return False
    
    @property
    def docstring(self) -> Optional[str]:
        if self.gxparam:
            return self.gxparam.docstring
        return ''
        #return f'examples: {", ".join(self.values.unique[:3])}'

    def update(self, incoming: Any) -> None:
        # transfer values
        assert(isinstance(incoming, Positional))
        self.values.record += incoming.values.record
        # transfer galaxy param reference
        if not self.gxparam and incoming.gxparam:
            self.gxparam: Optional[Param] = incoming.gxparam

    def __str__(self) -> str:
        return f'{str(self.default_value):20}{str(self.optional):>10}'




from galaxy2janis.logs import logging
from galaxy2janis import settings

from galaxy.tools import Tool as GxTool
from galaxy.tools.parameters.basic import ToolParameter
from galaxy.tool_util.parser.output_objects import ToolOutput

from ..interaction import get_tool
from ..configfiles.Configfile import Configfile

from .XMLToolDefinition import XMLToolDefinition
from .ToolXMLMetadata import ToolXMLMetadata
from .citations import Citation
from .parsing.ParamFlattener import ParamFlattener
from .param.ParamRegister import ParamRegister
from .TestRegister import TestRegister
from .requirements import CondaRequirement, ContainerRequirement

from .parsing.outputs import parse_output_param
from .parsing.inputs import parse_input_param


Requirement = ContainerRequirement | CondaRequirement


def load_xmltool() -> XMLToolDefinition:
    gxtool = get_tool(settings.tool.tool_path)
    factory = GalaxyToolFactory(gxtool)
    return factory.create()


class GalaxyToolFactory:
    def __init__(self, gxtool: GxTool):
        self.gxtool = gxtool
        self.inputs = self.parse_inputs()
        self.outputs = self.parse_outputs()

    def create(self) -> XMLToolDefinition:
        return XMLToolDefinition(
            self.parse_metadata(),
            self.parse_command(),
            self.parse_configfiles(),
            self.inputs,
            self.outputs,
            self.parse_tests()
        )
    
    def parse_inputs(self) -> ParamRegister:
        """returns a an InputRegister by reformatting the galaxy tool representation's params."""
        register = ParamRegister()
        g_in_params = self.flatten_params()
        for g_param in g_in_params:
            t_param = parse_input_param(g_param)
            register.add(t_param)
        return register

    def flatten_params(self) -> list[ToolParameter]:
        pf = ParamFlattener(self.gxtool.inputs)
        return pf.flatten()

    def parse_outputs(self) -> ParamRegister:
        """returns a formatted list of outputs using the representation"""
        register = ParamRegister()
        g_out_params: list[ToolOutput] = list(self.gxtool.outputs.values())
        for g_param in g_out_params:
            t_params = parse_output_param(g_param, self.inputs)
            for param in t_params:
                register.add(param)
        return register

    def parse_metadata(self, ) -> ToolXMLMetadata:
        """returns a formatted Metadata using the representation"""
        requirements: list[Requirement] = self.get_requirements()
        citations: list[Citation] = self.get_citations()
        return ToolXMLMetadata(
            name=str(self.gxtool.name),  #type: ignore
            id=str(self.gxtool.id),  #type: ignore
            version=str(self.gxtool.version).split('+galaxy')[0],  #type: ignore
            description=str(self.gxtool.description),  #type: ignore
            help=str(self.gxtool.raw_help),  #type: ignore
            requirements=requirements,
            citations=citations,
            creator=self.gxtool.creator  #type: ignore
        )
    
    def get_requirements(self) -> list[Requirement]:
        """returns a formatted list of Requirements using the representation"""
        reqs: list[Requirement] = []
        reqs += self.get_conda_requirements()
        reqs += self.get_container_requirements()
        return reqs
    
    def get_conda_requirements(self) -> list[CondaRequirement]:
        packages: list[dict[str, str]] = self.gxtool.requirements.to_list() # type: ignore
        return [CondaRequirement(p['name'], p['version']) for p in packages]

    def get_container_requirements(self) -> list[ContainerRequirement]:
        containers: list[ContainerDescription] = self.gxtool.containers # type: ignore
        return [ContainerRequirement(c.identifier) for c in containers] # type: ignore

    def get_citations(self) -> list[Citation]:
        citations: list[Citation] = []
        citations += self.get_biotools_citations()
        citations += self.get_doi_citations()
        #citations += self.get_bibtex_citations()
        return citations

    def get_biotools_citations(self) -> list[Citation]: 
        out: list[Citation] = []
        for ref in self.gxtool.xrefs:
            citation = Citation(
                type='biotools',
                text=f"https://bio.tools/{ref['value']}"
            )
            out.append(citation)
        return out

    def get_doi_citations(self) -> list[Citation]:
        out: list[Citation] = []
        for elem in self.gxtool.tool_source.xml_tree.findall('citations'):
            for ref in elem.findall('citation'):
                if ref.attrib['type'] == 'doi':
                    citation = Citation(
                        type='doi',
                        text=f"https://doi.org/{ref.text}"
                    )
                    out.append(citation)
        return out
        
    def get_bibtex_citations(self) -> list[Citation]:
        out: list[Citation] = []
        for elem in self.gxtool.tool_source.xml_tree.findall('citations'):
            for ref in elem.findall('citation'):
                if ref.attrib['type'] == 'bibtex':
                    citation = Citation(
                        type='bibtex',
                        text=self.parse_bibtex(ref.text)
                    )
                    out.append(citation)
        return out
    
    def parse_bibtex(self, bibtex_citation: dict[str, str]) -> str:
        # define and parse using biblib
        bp = bib.Parser()
        data = bp.parse(bibtex_citation).get_entries() # type: ignore
        # get each citation key: value pair
        entry = list(data.values())[0]  # type: ignore
        if 'url' in entry:
            return f"{entry['url']}" # type: ignore
        elif 'author' in entry and 'title' in entry:
            return f"{entry['author']}.  {entry['title']}"
        return ''

    def parse_command(self) -> str:
        """returns the tool xml command"""
        return str(self.gxtool.command) # type: ignore
    
    def parse_configfiles(self) -> list[Configfile]:
        """returns the tool configfiles"""
        out: list[Configfile] = []
        for name, _, contents in self.gxtool.config_files:  # type: ignore
            if isinstance(contents, str):
                new_config = Configfile(name, contents)  # type: ignore
                out.append(new_config)
        if out:
            logging.has_configfile()
        return out
    
    def parse_tests(self) -> TestRegister:
        """
        returns a formatted list of tests using the representation
        needs to be properly fleshed out later!
        """
        return TestRegister([])






import json
import requests
from typing import Any, Optional
from requests import Response

from galaxy2janis.logs import logging



def make_api_request(request_url: str) -> Any:
    response = make_request(request_url)
    return handle_response(response)

def make_request(request_url: str) -> Optional[Response]:
    # make requests to get information about tools with similar name
    iterations = 1
    response: Optional[Response] = None
    while response is None:
        if iterations > 5:
            break
        try:
            response = requests.get(request_url, timeout=5)
        except requests.exceptions.Timeout:
            response = None
        iterations += 1
    return response

def handle_response(response: Optional[Response]) -> Any:
    if response is None or response.status_code != 200:
        logging.no_ga4gh_data()
        return None
    else:
        return json.loads(response.text)




from abc import ABC, abstractmethod
from typing import Any, Optional

from galaxy2janis import expressions
from galaxy2janis.expressions.patterns import COMPOUND_OPT

from ...gxtool.param.Param import Param
from ..tokens import Token
from ..tokens import TokenType
from ..components import Tee
from ..components import StreamMerge
from ..components import factory
from .ExecutionPath import EPathPosition
from . import utils as component_utils


class Annotator(ABC):

    """extracts a command component at the current position in an epath"""
    def __init__(self, ptr: int, positions: list[EPathPosition]):
        self.ptr = ptr
        self.positions = positions

        self.ctoken: Token = self.positions[self.ptr].token
        self.ntoken: Token = self.positions[self.ptr + 1].token
        self.success: bool = False

    def annotate(self) -> None:
        if self.passes_check():
            self.handle()
            self.success = True

    def update_epath_components(self, ptr: int, component: Any) -> None:
        self.positions[ptr].component = component
        self.positions[ptr].ignore = True

    @abstractmethod
    def passes_check(self) -> bool:
        """confirm whether this type of component should be extracted at the current position"""
        ...
    
    @abstractmethod
    def handle(self) -> None:
        """perform the necessary operations now that the component is valid"""
        ...
   
    @abstractmethod
    def calculate_next_ptr_pos(self) -> int:
        """update the ptr pos to continue iteration. based on component span"""
        ...
    

# --------------
# TOOL ARGUMENTS
# --------------

class OptionAnnotator(Annotator):

    """this one is complex, others are straightforward"""
    def __init__(self, ptr: int, positions: list[EPathPosition]):
        super().__init__(ptr, positions)
        self.stop_ptr: int = 0

    def passes_check(self) -> bool:
        if component_utils.is_option(self.ctoken, self.ntoken):
            if not component_utils.has_compound_structure(self.ctoken):
                return True
        return False
    
    def handle(self) -> None:
        value_tokens = self.get_option_values()
        gxparam = self.get_param(value_tokens)
        option = factory.option(
            prefix=self.ctoken.text,
            gxparam=gxparam,
            delim=self.get_delim(),
            values=[v.text for v in value_tokens]
        )
        for pos in range(self.ptr, self.stop_ptr + 1):
            self.update_epath_components(pos, option)

    def get_param(self, value_tokens: list[Token]) -> Optional[Param]:
        out: Optional[Param] = self.ctoken.gxparam
        if not out:
            for token in value_tokens:
                if token.gxparam:
                    out = token.gxparam
                    break
        return out

    def get_delim(self) -> str:
        if self.ntoken.ttype == TokenType.KV_LINKER:
            return self.ntoken.text
        return ' ' # fallback default

    def get_option_values(self) -> list[Token]:
        if self.is_kv_pair():
            return self.get_single_value()
        else:
            return self.get_multiple_values()

    def is_kv_pair(self) -> bool:
        if self.positions[self.ptr + 1].token.ttype == TokenType.KV_LINKER:
            return True
        return False
     
    def get_single_value(self) -> list[Token]:
        # if is_kv_pair() -- self.ptr + 1 will always be KV_LINKER
        self.stop_ptr = self.ptr + 2
        value = self.positions[self.ptr + 2]
        return [value.token]

    def get_multiple_values(self) -> list[Token]:
        out: list[Token] = []

        # define a max possible end point to consume upto (handles edge cases)
        final_pos = self.get_greedy_consumption_end()  
        values_type = self.positions[self.ptr + 1].token.ttype

        # look at next ntoken to see its probably a value for the option
        curr_pos = self.ptr + 1
        while curr_pos <= final_pos:
            if self.should_eat_value(curr_pos, values_type):
                out.append(self.positions[curr_pos].token)
                curr_pos += 1 
            else:
                break
        self.stop_ptr = curr_pos - 1
        return out

    def should_eat_value(self, curr_pos: int, values_type: TokenType) -> bool:
        ctoken = self.positions[curr_pos].token
        ntoken = self.positions[curr_pos + 1].token
        if component_utils.is_positional(ctoken) and ctoken.ttype == values_type and ntoken.ttype != TokenType.KV_LINKER:
            return True
        return False

    def get_greedy_consumption_end(self) -> int:
        """
        define an end point -> tee or redirect or end of positions
        most cases: end point = len(positions) to the end, first tee, first redirect etc - 2
        special case: 2nd last position. --files $input1 set end point = the above - 1
        """
        final_pos = self.get_end_token_position()
        if self.ptr < final_pos - 1:
            final_pos -= 1
        return final_pos

    def get_end_token_position(self) -> int:
        end_tokens = [TokenType.END_STATEMENT, TokenType.LINUX_TEE, TokenType.LINUX_REDIRECT]
        for position in self.positions:
            if position.token.ttype in end_tokens:
                return position.ptr - 1
        return len(self.positions) - 2  # this will never happen
    
    def calculate_next_ptr_pos(self) -> int:
        return self.stop_ptr + 1


class CompoundOptionAnnotator(Annotator):

    def passes_check(self) -> bool:
        if component_utils.is_option(self.ctoken, self.ntoken):
            if component_utils.has_compound_structure(self.ctoken):
                return True
        return False
    
    def handle(self) -> None:
        match = expressions.get_matches(self.ctoken.text, COMPOUND_OPT)[0]
        option = factory.option(match.group(1), gxparam=self.ctoken.gxparam, delim='', values=[match.group(2)])
        self.update_epath_components(self.ptr, option)
    
    def calculate_next_ptr_pos(self) -> int:
        return self.ptr + 1


class FlagAnnotator(Annotator):

    def passes_check(self) -> bool:
        if component_utils.is_flag(self.ctoken, self.ntoken):
            return True
        return False
    
    def handle(self) -> None:
        prefix = self.ctoken.text
        gxparam = self.ctoken.gxparam
        flag = factory.flag(prefix, gxparam)
        self.update_epath_components(self.ptr, flag)

    def calculate_next_ptr_pos(self) -> int:
        return self.ptr + 1


class PositionalAnnotator(Annotator):

    def passes_check(self) -> bool:
        if component_utils.is_positional(self.ctoken):
            return True
        return False
    
    def handle(self) -> None:
        value = self.ctoken.text
        gxparam = self.ctoken.gxparam
        positional = factory.positional(value, gxparam)
        self.update_epath_components(self.ptr, positional)
    
    def calculate_next_ptr_pos(self) -> int:
        return self.ptr + 1
 

# --------------
# LINUX CONSTRUCTS
# --------------

class StreamMergeAnnotator(Annotator):

    def passes_check(self) -> bool:
        if self.ctoken.ttype == TokenType.LINUX_STREAM_MERGE:
            return True
        return False
    
    def handle(self) -> None:
        component = StreamMerge(self.ctoken.text)
        self.update_epath_components(self.ptr, component)
   
    def calculate_next_ptr_pos(self) -> int:
        return self.ptr + 1


class RedirectAnnotator(Annotator):

    def passes_check(self) -> bool:
        if self.ctoken.ttype == TokenType.LINUX_REDIRECT:
            return True
        return False
    
    def handle(self) -> None:
        if self.ctoken and self.ntoken:
            component = factory.redirect_output(self.ctoken.text, self.ntoken.text, gxparam=self.ntoken.gxparam)
            self.update_epath_components(self.ptr, component)
            self.update_epath_components(self.ptr + 1, component)
   
    def calculate_next_ptr_pos(self) -> int:
        return self.ptr + 2


class TeeAnnotator(Annotator):

    def passes_check(self) -> bool:
        if self.ctoken.ttype == TokenType.LINUX_TEE:
            return True
        return False

    def handle(self) -> None:
        tee = Tee()
        ptr = self.ptr + 1 # TODO debug -> does this change self.ptr?

        # tee arguments
        token = self.positions[ptr].token
        while token.text.startswith('-'):
            tee.options.append(token.text)
            self.update_epath_components(ptr, tee)
            ptr += 1
            token = self.positions[ptr].token

        # tee files: consumes to end of statement
        while ptr < len(self.positions) - 1:
            token = self.positions[ptr].token
            tee.files.append(token.text)
            self.update_epath_components(ptr, tee)
            ptr += 1

    def calculate_next_ptr_pos(self) -> int:
        return len(self.positions)








from typing import Any, Optional
from galaxy2janis.gx.command.cmdstr.DynamicCommandStatement import DynamicCommandStatement
from galaxy2janis.gx.command.cmdstr.generate import gen_command_string
from galaxy2janis.entities.workflow.workflow import Workflow
from galaxy2janis.gx.gxtool import load_xmltool
from galaxy2janis.gx.gxtool.text import load_partial_cheetah_command

from galaxy2janis import mapping
from galaxy2janis import settings


def ingest_workflow_steps_prepost(janis: Workflow, galaxy: dict[str, Any]) -> None:
    for g_step in galaxy['steps'].values():
        if g_step['type'] == 'tool':
            ingest_prepost(g_step, janis, galaxy)

def ingest_prepost(g_step: dict[str, Any], janis: Workflow, galaxy: dict[str, Any]):
    # get janis step & update settings
    j_step = mapping.step(g_step['id'], janis, galaxy)
    settings.tool.set(wrapper=j_step.metadata.wrapper)

    xmltool = load_xmltool()
    command = load_partial_cheetah_command(inputs_dict=g_step['tool_state'])
    cmdstr = gen_command_string(source='xml', the_string=command, xmltool=xmltool)
    j_step.preprocessing = extract_cmdline(cmdstr.preprocessing)
    j_step.postprocessing = extract_cmdline(cmdstr.postprocessing)

def extract_cmdline(statements: list[DynamicCommandStatement]) -> Optional[str]:
    if not statements:
        return None
    cmdlines = [x.cmdline for x in statements]
    return ' &&'.join(cmdlines)







from __future__ import annotations
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from galaxy2janis.entities.workflow import Workflow

from galaxy2janis.entities.workflow import WorkflowInput
from galaxy2janis.entities.workflow import ConnectionInputValue
from galaxy2janis.entities.workflow import WorkflowInputInputValue

from galaxy2janis import mapping
from galaxy2janis import settings


def handle_tool_connection_inputs(janis: Workflow, galaxy: dict[str, Any]) -> None:
    """connections listed in step 'input_connections'"""
    ingestor = ConnectionInputIngestor(janis, galaxy)
    ingestor.ingest()


class ConnectionInputIngestor:
    def __init__(self, janis: Workflow, galaxy: dict[str, Any]):
        self.janis = janis
        self.galaxy = galaxy

    def ingest(self) -> None:
        for step in self.galaxy['steps'].values():
            if step['type'] == 'tool':
                self.ingest_connections(step)

    def ingest_connections(self, g_step: dict[str, Any]) -> None:
        """
        ingest connections for this galaxy step.
        \n
        for each connection, need to know the emitting entity and the target entity.
        the emitting entity is a step output, and the target is a tool component.
        there isn't always a 1-to-1 mapping here. this is due to some galaxy params not being
        linked to a tool component. 
        
        j_emitter:  StepOutput or WorkflowInput (the source of the connection)
        j_target: a Tool InputComponent (the destination of the connection)

        """
        j_step = mapping.step(g_step['id'], self.janis, self.galaxy)
        settings.tool.set(wrapper=j_step.metadata.wrapper)
        for g_target, g_emitter in g_step['input_connections'].items():
            g_target = g_target.replace('|', '.')
            j_target = mapping.tool_input(g_target, j_step.tool)
            j_emitter = mapping.emitter(g_emitter['id'], g_emitter['output_name'], self.janis, self.galaxy)

            if isinstance(j_emitter, WorkflowInput):
                value = WorkflowInputInputValue(
                    component=j_target, # j_target can be InputComponent or None
                    input_uuid=j_emitter.uuid,
                    is_runtime=False
                )
            else:
                j_emitter_step = mapping.step(g_emitter['id'], self.janis, self.galaxy)
                value = ConnectionInputValue(
                    component=j_target,
                    step_uuid=j_emitter_step.uuid,
                    out_uuid=j_emitter.tool_output.uuid
                )

            j_step.inputs.add(value)




from __future__ import annotations
from typing import Optional

from .OutputComponent import OutputComponent
from ..linux.streams import Stream
from ...components.ValueRecord import ValueRecord


class RedirectOutput(OutputComponent):
    def __init__(self, redirect: str, filepath: str):
        super().__init__()
        self.redirect = redirect
        self.stream: Stream = self.extract_stream()
        self.values: ValueRecord = ValueRecord()
        self.values.add(filepath)

    @property
    def name(self) -> str:
        name: str = 'redirect'
        # get name from galaxy param if available
        if self.gxparam:
            name = self.gxparam.name
        # otherwise, most commonly witnessed option value as name
        elif self.values.most_common_value:
            name = self.values.most_common_value
        if not name.startswith('out'):
            name = f'out_{name}'
        return name
    
    @property
    def text(self) -> str:
        filepath = self.values.most_common_value
        assert(filepath)
        return f'{self.redirect} {filepath}'

    @property
    def optional(self) -> bool:
        # NOTE - janis does not allow optional outputs
        return False

    @property
    def docstring(self) -> Optional[str]:
        if self.gxparam:
            return self.gxparam.docstring
        return ''
    
    def is_append(self) -> bool:
        if '>>' in self.redirect:
            return True
        return False

    def extract_stream(self) -> Stream:
        if self.redirect[0] == '2':
            return Stream.STDERR
        return Stream.STDOUT







import os
import json
import requests
from typing import Any
from bs4 import BeautifulSoup, Tag

# FILE IO 


def load_data(path: str) -> dict[str, Any]:
    """lockfile?"""
    if not os.path.exists(path):
        return {}
    with open(path, 'r') as fp:
        return json.load(fp)

def save_data(path: str, data: dict[str, Any]) -> None:
    """lockfile?"""
    with open(path, 'w') as fp:
        json.dump(data, fp)


# REQUESTS / HTML PARSING
def get_soup(url: str) -> BeautifulSoup:
    response = requests.get(url)
    return BeautifulSoup(response.content, "html.parser")

def get_table(page: BeautifulSoup) -> Tag:
    table = page.find(name='table')
    if not isinstance(table, Tag):
        raise RuntimeError()
    return table

def get_table_col(table: Tag, col: int) -> list[Tag]:
    rows = table.find_all('tr')
    rows = [row for row in rows if row.parent.name == 'tbody']
    elems = [row.find_all('td')[col] for row in rows]
    return elems

def format_iframe_url(iframe_elem: Tag) -> str:
    return f'https://toolshed.g2.bx.psu.edu/{iframe_elem.attrs["src"]}'







from typing import Optional
from uuid import uuid4
from galaxy2janis import tags

from .step.outputs import StepOutput
from .step.step import WorkflowStep
from .metadata import WorkflowMetadata
from .input import WorkflowInput



class Workflow:
    """
    a Workflow() is the final representation of the galaxy workflow
    being parsed. Includes workflow metadata, inputs, tool steps, and outputs. 
    step_ids are used to lookup key information. 
    each input, step, and output is stored with step_id as key. 
    that way everything can be referenced properly. 
    """
    def __init__(self):
        self.uuid: str = str(uuid4())
        self.inputs: list[WorkflowInput] = []
        self._steps: list[WorkflowStep] = []
        self._metadata: Optional[WorkflowMetadata] = None

    @property
    def name(self) -> str:
        return self.metadata.name
    
    @property
    def tag(self) -> str:
        return tags.get(self.uuid)
    
    @property
    def steps(self) -> list[WorkflowStep]:
        the_list = self._steps
        the_list.sort(key=lambda x: x.metadata.step_id)
        return the_list

    @property
    def metadata(self) -> WorkflowMetadata:
        if self._metadata:
            return self._metadata
        raise RuntimeError('no metadata set')

    @property
    def outputs(self) -> list[StepOutput]:
        workflow_outputs: list[StepOutput] = []
        for step in self.steps:
            for out in step.outputs.list():
                if out.is_wflow_out:
                    workflow_outputs.append(out)
        return workflow_outputs

    def set_metadata(self, metadata: WorkflowMetadata) -> None:
        self._metadata = metadata
        tags.new_group('workflow', self.uuid)
        tags.register(self)
    
    def add_input(self, w_inp: WorkflowInput) -> None:
        tags.switch_group(self.uuid)
        tags.register(w_inp)
        self.inputs.append(w_inp)

    def add_step(self, step: WorkflowStep) -> None:
        tags.switch_group(self.uuid)
        tags.register(step)
        self.steps.append(step)

    def get_input(self, query_uuid: str) -> WorkflowInput:
        for winp in self.inputs:
            if winp.uuid == query_uuid:
                return winp
        raise RuntimeError('could not find input with uuid')
    
    def get_input_children(self, query_uuid: str) -> list[WorkflowStep]:
        out: list[WorkflowStep] = []
        for step in self.steps:
            for winp_connection in step.inputs.workflow_inputs:
                if winp_connection.input_uuid == query_uuid:
                    out.append(step)
        return out
    
    def get_step(self, query_uuid: str) -> WorkflowStep:
        for step in self.steps:
            if step.uuid == query_uuid:
                return step
        raise RuntimeError('could not find step with uuid')
    
    def get_step_children(self, query_uuid: str) -> list[WorkflowStep]:
        out: list[WorkflowStep] = []
        for step in self.steps:
            for connection in step.inputs.connections:
                if connection.step_uuid == query_uuid:
                    out.append(step)
        return out
    
    def get_step_output(self, query_uuid: str) -> StepOutput:
        for step in self.steps:
            for s_out in step.outputs.list():
                if s_out.tool_output.uuid == query_uuid:
                    return s_out
        raise RuntimeError('could not find step output with uuid')
    







from dataclasses import dataclass
from typing import Any, Optional

from galaxy2janis import expressions


@dataclass
class Version:
    text: str

    def get_numeric(self) -> str:
        matches = expressions.get_matches(self.text, expressions.patterns.VERSION)
        if matches:
            match = matches[0]
            return match[0] # type: ignore
        raise RuntimeError()

    def get_numeric_levels(self, fill_size: int=0) -> list[int]:
        numeric_version = self.get_numeric()
        split_version = numeric_version.split('.')
        split_version = [int(x) for x in split_version]
        while len(split_version) < fill_size:
            split_version.append(0)
        return split_version
        

class VersionMatcher:

    def get_version_exact(self, tool_data: dict[str, Any], target_version: str) -> Optional[dict[str, str]]:
        for ver in tool_data['versions']:
            if ver['meta_version'] == target_version:
                return ver
        return None

    def get_version_trimmed(self, tool_data: dict[str, Any], target_version: str) -> Optional[dict[str, str]]:
        target = Version(target_version)

        for ver in tool_data['versions']:
            query = Version(ver['meta_version'])
            if query.get_numeric():
                if query.get_numeric() == target.get_numeric():
                    return ver
        return None

    def get_version_trimmed_inexact(self, tool_data: dict[str, Any], target_version: str) -> Optional[dict[str, str]]:
        """
        selects the version by trimming to numeric, then matching most similar, allowing wobble 
        just does a version comparision sort for each digit, right to left. 
        sorts are stable, so ordering comes out correct
        """
        target = Version(target_version)
        queries = [Version(ver['meta_version']) for ver in tool_data['versions']]
        array_size = max([len(ver.get_numeric_levels()) for ver in [target] + queries])

        for i in range(array_size -1, -1, -1):
            queries.sort(key=lambda x: abs(target.get_numeric_levels()[i] - x.get_numeric_levels(fill_size=array_size)[i]))
        
        selected = queries[0].text
        #return the right version dict
        for query in tool_data['versions']:
            if query['meta_version'] == selected:
                return query

    def get_most_recent(self, tool_data: dict[str, Any]) -> dict[str, str]:
        return tool_data['versions'][-1]
        